
/// Type alias to represent a register.
type RegisterMem = String;

/// Type alias to represent a label used in place of a const.
pub type ConstId = String;

/// A const can either be a number or a label of some section.
#[derive(Debug)]
pub enum Const {
    CS(ConstId),
    C(u8),
}

/// Operations as they are parsed, before translating them to binary.
/// This type is used internally by the parser and 
/// differs from cpu::Operation.
#[derive(Debug)]
pub enum Operation {
    /// No type.
    NOP,
    HALT,
    /// R type
    ADD(RegisterMem, RegisterMem, RegisterMem),
    SUB(RegisterMem, RegisterMem, RegisterMem),
    AND(RegisterMem, RegisterMem, RegisterMem),
    XOR(RegisterMem, RegisterMem, RegisterMem),
    SLL(RegisterMem, RegisterMem, RegisterMem),
    BEQ(RegisterMem, RegisterMem, RegisterMem),
    BGT(RegisterMem, RegisterMem, RegisterMem),
    LOAD(RegisterMem, RegisterMem, RegisterMem),
    STORE(RegisterMem, RegisterMem, RegisterMem),

    /// I Type
    SLI(RegisterMem, Const),
    ADDI(RegisterMem, Const),
    CALL(RegisterMem, Const),

    /// J Type
    JAL(RegisterMem, RegisterMem, Const),
}


pub fn print_op(op: Operation) -> String {

    let print_const = |x : Const| {
        match x {
            Const::CS(i) => i,
            Const::C(n) => format!("{}", n),
        }
    };

    match op {
        Operation::NOP => String::from("nop"),
        Operation::HALT => String::from("halt"),
        Operation::ADD(a, b, c) => format!("add {} {} {}", a, b, c),
        Operation::SUB(a, b, c) => format!("sub {} {} {}", a, b, c),
        Operation::AND(a, b, c) => format!("and {} {} {}", a, b, c),
        Operation::XOR(a, b, c) => format!("xor {} {} {}", a, b, c),
        Operation::SLL(a, b, c) => format!("sll {} {} {}", a, b, c),
        Operation::BEQ(a, b, c) => format!("beq {} {} {}", a, b, c),
        Operation::BGT(a, b, c) => format!("bgq {} {} {}", a, b, c),
        Operation::LOAD(a, b, c) => format!("load {} {} {}", a, b, c),
        Operation::STORE(a, b, c) => format!("store {} {} {}", a, b, c),
        Operation::SLI(a, c) => format!("sli {} {}", a, print_const(c)),
        Operation::ADDI(a, c) => format!("addi {} {}", a, print_const(c)),
        Operation::CALL(a, c) => format!("call {} {}", a, print_const(c)),
        Operation::JAL(a, b, c) => format!("jal {} {} {}", a, b, print_const(c)),
    }

}

