use super::AST::*;
use crate::cpu::get_num;

use log::{trace, warn};


/// Trait to represent a format we can translate our assembly to.
pub trait CodeFormat {
    fn encode_op(op: &Operation, sy: &SymbolTable, current_pc: u16) -> Option<Self>
    where
        Self: Sized;
}

/// Symbol table implemented as a vector.
/// This is a zero-size struct used to implement
/// lookup as an impl.

#[repr(transparent)]
#[derive(Debug)]
pub struct SymbolTable(pub Vec<(String, u16)>);

impl SymbolTable {
    /// Finds a symbol in the symbol table.
    /// Fails if the symbol is not in the symbol table.
    pub fn lookup(&self, query: &str) -> Option<u16> {
        trace!("Looking up {} in the symbol table.", query);
        let SymbolTable(sy) = self;

        for (name, loc) in sy.into_iter() {
            if query == (*name) {
                return Some(*loc);
            }
        }
        // panic!(
        //     "Symbol {} not found in symbol table. \nCurrent symbol table:{:?}",
        //     query, sy
        // );
        warn!("Symbol {} not found in symbol table.", query);
        return None
    }
}

impl CodeFormat for u16 {
    fn encode_op(op: &Operation, sy: &SymbolTable, current_pc: u16) -> Option<Self> {
        trace!("encoding {:?}", op);
        match op {
            Operation::NOP => Some(0b0000000000000000),
            Operation::HALT => Some(0b1111111111111111),
            Operation::ADD(r1, r2, r3) => {
                let (r1b, r2b, r3b) = (
                    get_num(&r1)? as u16,
                    get_num(&r2)? as u16,
                    get_num(&r3)? as u16,
                );
                return Some((0b0001 << 12) + (r1b << 8) + (r2b << 4) + r3b);
            }
            Operation::SUB(r1, r2, r3) => {
                let (r1b, r2b, r3b) = (
                    get_num(&r1)? as u16,
                    get_num(&r2)? as u16,
                    get_num(&r3)? as u16,
                );
                return Some((0b0010 << 12) + (r1b << 8) + (r2b << 4) + r3b);
            }
            Operation::AND(r1, r2, r3) => {
                let (r1b, r2b, r3b) = (
                    get_num(&r1)? as u16,
                    get_num(&r2)? as u16,
                    get_num(&r3)? as u16,
                );
                return Some((0b0011 << 12) + (r1b << 8) + (r2b << 4) + r3b);
            }
            Operation::XOR(r1, r2, r3) => {
                let (r1b, r2b, r3b) = (
                    get_num(&r1)? as u16,
                    get_num(&r2)? as u16,
                    get_num(&r3)? as u16,
                );
                return Some((0b0100 << 12) + (r1b << 8) + (r2b << 4) + r3b);
            }
            Operation::SLL(r1, r2, r3) => {
                let (r1b, r2b, r3b) = (
                    get_num(&r1)? as u16,
                    get_num(&r2)? as u16,
                    get_num(&r3)? as u16,
                );
                return Some((0b0101 << 12) + (r1b << 8) + (r2b << 4) + r3b);
            }
            Operation::BEQ(r1, r2, r3) => {
                let (r1b, r2b, r3b) = (
                    get_num(&r1)? as u16,
                    get_num(&r2)? as u16,
                    get_num(&r3)? as u16,
                );
                return Some((0b1000 << 12) + (r1b << 8) + (r2b << 4) + r3b);
            }
            Operation::BGT(r1, r2, r3) => {
                let (r1b, r2b, r3b) = (
                    get_num(&r1)? as u16,
                    get_num(&r2)? as u16,
                    get_num(&r3)? as u16,
                );
                return Some((0b1001 << 12) + (r1b << 8) + (r2b << 4) + r3b);
            }
            Operation::LOAD(r1, r2, r3) => {
                let (r1b, r2b, r3b) = (
                    get_num(&r1)? as u16,
                    get_num(&r2)? as u16,
                    get_num(&r3)? as u16,
                );
                return Some((0b1100 << 12) + (r1b << 8) + (r2b << 4) + r3b);
            }
            Operation::STORE(r1, r2, r3) => {
                let (r1b, r2b, r3b) = (
                    get_num(&r1)? as u16,
                    get_num(&r2)? as u16,
                    get_num(&r3)? as u16,
                );
                return Some((0b1101 << 12) + (r1b << 8) + (r2b << 4) + r3b);
            }
            Operation::SLI(r1, c) => {
                let r1b = get_num(&r1)? as u16;
                let cb = match c {
                    Const::CS(label) => sy.lookup(&label)?,
                    Const::C(n) => (*n) as u16,
                };
                return Some((0b0110 << 12) + (r1b << 8) + cb);
            }
            Operation::ADDI(r1, c) => {
                let r1b = get_num(&r1)? as u16;
                let cb = match c {
                    Const::CS(label) => sy.lookup(&label)?,
                    Const::C(n) => (*n) as u16,
                };
                return Some((0b0111 << 12) + (r1b << 8) + cb);
            }
            Operation::CALL(r1, c) => {
                let r1b = get_num(&r1)? as u16;
                let cb = match c {
                    Const::CS(label) => sy.lookup(&label)?,
                    Const::C(n) => (*n) as u16,
                };
                return Some((0b1110 << 12) + (r1b << 8) + cb);
            }
            Operation::JAL(r1, r2, c) => {
                let r1b = get_num(&r1)? as u16;
                let r2b = get_num(&r2)? as u16;
                let cb = match c {
                    Const::CS(label) => current_pc - sy.lookup(&label)?,
                    Const::C(n) => (*n) as u16,
                };
                return Some((0b1010 << 12) + (r1b << 8) + (r2b << 4) + cb);
            }
        }
    }
}
