mod AST;
pub mod encoder;
pub mod parser;
mod tests;

use encoder::CodeFormat;
use encoder::SymbolTable;

pub use AST::print_op;

use log::{trace, debug};

use parser::Section;

use crate::loader::unloader::make_string;

use parser::SectionContent;

impl Section {
    /// Calculates the size, in binary, of a section.
    fn get_size(&self) -> usize {
        match &self.content {
            // code is 1 word for instruction.
            SectionContent::Code(c) => c.len(),

            // UTF-8 strings are collections of 8-bit chunks, but they're
            // packed into words of 16 bits + 16 bit NULL.
            //
            // If there's an uneven number of bytes in a string, 
            // we add a 8 bit empty padding and then the NULL byte.
            SectionContent::CString(s) => {
                let c = s.len();
                if c % 2 != 0 {
                    (c + 1) / 2 + 1
                } else {
                    c / 2 + 1
                }
            }
            SectionContent::CVec() => todo!(),
        }
    }

    /// Converts this section to binary. Needs symbol table to
    /// resolve labels, and to quickly get the address of 
    /// this own section.
    fn to_binary(&self, sy: &SymbolTable) -> Option<crate::loader::Section> {
        let own_address = sy.lookup(&self.name)?;
        match &self.content {
            SectionContent::Code(c) => {
                let mut res = vec![];

                // we keep track of the program counter because we
                // need to calculate relative jumps.
                let mut pc = own_address;
                for op in c.iter() {
                    trace!("converting {:?}", op);

                    res.push(CodeFormat::encode_op(op, sy, pc)?);
                    // pc simply increases by one after each operation.
                    pc += 1;
                }

                return Some(crate::loader::Section::new(self.name.clone(), &res));
            }
            SectionContent::CString(s) => {
                return Some(crate::loader::Section::new(self.name.clone(), &make_string(s)));
            }
            SectionContent::CVec() => todo!(),
        }
    }
}


/// Sorts a list of sections.
/// All .text sections containing code are
/// put at the beginning of the binary file, in the order they 
/// appear in the assembly file, except the "main" section,
/// which is the entrypoint of our program and must be put
/// at the very beginning.
fn sort_sections(sections: Vec<Section>) -> Option<Vec<Section>> {
    // we start with a mock section that we'll just replace.
    let mut res: Vec<Section> = vec![Section {
        name: "".to_owned(),
        content: SectionContent::CString("".to_owned()),
    }];

    let mut nocode: Vec<Section> = vec![];

    for section in sections.into_iter() {
        // trick: if let to pattern match discarding result
        if let SectionContent::Code(_) = section.content {
            if section.name == "main" {
                res[0] = section;
            } else {
                res.push(section);
            }
        } else {
            nocode.push(section);
        }
    }

    res.append(&mut nocode);

    // TODO: PANIC WHEN NO MAIN;

    return Some(res);
}

/// Creates symbol from a list of sorted sections.
/// Assumes the sections are already sorted in the
/// desired order.
fn make_symbol_table<'a>(sections: &'a Vec<Section>) -> Option<SymbolTable> {
    let mut res = vec![];
    let mut pos: u16 = 0;

    for sec in sections {
        res.push((sec.name.clone(), pos));
        pos += sec.get_size() as u16;
    }

    return Some(SymbolTable(res));
}

/// Converts a vector of sections into binary.
pub fn to_binary(sections: Vec<Section>) -> Option<Vec<crate::loader::Section>> {
    let sorted = sort_sections(sections)?;
    trace!("sorted sections: {:?}", sorted);
    let sy = make_symbol_table(&sorted)?;
    debug!("symbol table: {:?}", sy);
    let k = sorted
        .iter()
        .map(|x| x.to_binary(&sy))
        .collect::<Option<Vec<crate::loader::Section>>>()?;
    trace!("binary sections: {:?}", k);

    return Some(k);
}
