pub type Word = u16;
pub type Register = u8;

/// We need to hold 15 registers (zero is constant) + the program counter.
/// We'll just use a vector of u16.
#[derive(Debug)]
pub struct Registers {
    regs: [Word; 15],
    pub pc: Word,
}

impl Default for Registers {
    fn default() -> Self {
        Registers {
            regs: [0; 15],
            pc: 0,
        }
    }
}

impl Registers {
    /// retrives the register's value. Returns None if trying to access a register
    /// that doesn't exist.
    pub fn get(&self, i: Register) -> Option<Word> {
        match i {
            0 => Some(0), // zero is always 0
            1..=15 => Some(self.regs[(i - 1) as usize]),
            _ => None,
        }
    }

    /// writes val to the register i. Returns none on trying to write to zero, or to a register
    /// that doesn't exist.
    pub fn write(&mut self, i: Register, val: Word) -> Option<()> {
        match i {
            0 => None, // cannot write to 0
            1..=15 => {
                self.regs[(i - 1) as usize] = val;
                return Some(());
            }
            _ => None,
        }
    }
}

const ASSOCS: &'static [(Register, &'static str)] = &[
    (0, "zero"),
    (1, "ra"),
    (2, "sp"),
    (3, "t0"),
    (4, "t1"),
    (5, "t2"),
    (6, "t3"),
    (7, "a0"),
    (8, "a1"),
    (9, "a2"),
    (10, "a3"),
    (11, "s0"),
    (12, "s1"),
    (13, "s2"),
    (14, "s3"),
    (15, "t4"),
];

/// gets the register memonic name. Useful for pretty printing. (11 -> s0)
pub fn get_memo(i: Register) -> Option<&'static str> {
    for (a, b) in ASSOCS {
        if i == *a {
            return Some(b);
        }
    }

    return None;
}

/// gets the register index from its memonic name (s0 -> 11)
pub fn get_num(s: &str) -> Option<Register> {
    for (a, b) in ASSOCS {
        if s == *b {
            return Some(*a);
        }
    }

    return None;
}
